@@ -11,6 +11,7 @@
 # under the License.
 
 import typing as ty
+import warnings
 
 from openstack.block_storage import _base_proxy
 from openstack.block_storage.v3 import attachment as _attachment
@@ -36,6 +37,7 @@ from openstack import exceptions
 from openstack.identity.v3 import project as _project
 from openstack import resource
 from openstack import utils
+from openstack import warnings as os_warnings
 
 
 class Proxy(_base_proxy.BaseBlockStorageProxy):
@@ -68,7 +70,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
             instance.
 
         :returns: One :class:`~openstack.block_storage.v3.snapshot.Snapshot`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         """
         return self._get(_snapshot.Snapshot, snapshot)
@@ -85,7 +87,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
 
         :param snapshot: The name or ID a snapshot
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the snapshot does not exist. When set to ``True``, None will
             be returned when attempting to find a nonexistent resource.
         :param bool details: When set to ``False`` :class:
@@ -97,7 +99,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
             a higher chance of duplicates. Admin-only by default.
 
         :returns: One :class:`~openstack.block_storage.v3.snapshot.Snapshot`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         :raises: :class:`~openstack.exceptions.DuplicateResource` when multiple
             resources are found.
@@ -169,7 +171,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
         :param snapshot: The value can be either the ID of a snapshot or a
             :class:`~openstack.block_storage.v3.snapshot.Snapshot` instance.
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be
+            :class:`~openstack.exceptions.NotFoundException` will be
             raised when the snapshot does not exist.
             When set to ``True``, no exception will be set when
             attempting to delete a nonexistent snapshot.
@@ -291,7 +293,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
             :class:`~openstack.block_storage.v3.type.Type` instance.
 
         :returns: One :class:`~openstack.block_storage.v3.type.Type`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         """
         return self._get(_type.Type, type)
@@ -301,11 +303,11 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
 
         :param snapshot: The name or ID a volume type
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the volume type does not exist.
 
         :returns: One :class:`~openstack.block_storage.v3.type.Type`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         :raises: :class:`~openstack.exceptions.DuplicateResource` when multiple
             resources are found.
@@ -341,7 +343,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
         :param type: The value can be either the ID of a type or a
             :class:`~openstack.block_storage.v3.type.Type` instance.
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be
+            :class:`~openstack.exceptions.NotFoundException` will be
             raised when the type does not exist.
             When set to ``True``, no exception will be set when
             attempting to delete a nonexistent type.
@@ -436,7 +438,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
             instance.
 
         :returns: One :class:`~openstack.block_storage.v3.type.TypeEncryption`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         """
         volume_type = self._get_resource(_type.Type, volume_type_id)
@@ -482,7 +484,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
             instance.  Required if encryption_id is None.
 
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be
+            :class:`~openstack.exceptions.NotFoundException` will be
             raised when the type does not exist.
             When set to ``True``, no exception will be set when
             attempting to delete a nonexistent type.
@@ -541,7 +543,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
             :class:`~openstack.block_storage.v3.volume.Volume` instance.
 
         :returns: One :class:`~openstack.block_storage.v3.volume.Volume`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         """
         return self._get(_volume.Volume, volume)
@@ -558,7 +560,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
 
         :param snapshot: The name or ID a volume
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the volume does not exist.
         :param bool details: When set to ``False`` no extended attributes
             will be returned. The default, ``True``, will cause objects with
@@ -568,7 +570,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
             a higher chance of duplicates. Admin-only by default.
 
         :returns: One :class:`~openstack.block_storage.v3.volume.Volume`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         :raises: :class:`~openstack.exceptions.DuplicateResource` when multiple
             resources are found.
@@ -625,7 +627,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
         :param volume: The value can be either the ID of a volume or a
             :class:`~openstack.block_storage.v3.volume.Volume` instance.
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be
+            :class:`~openstack.exceptions.NotFoundException` will be
             raised when the volume does not exist.
             When set to ``True``, no exception will be set when
             attempting to delete a nonexistent volume.
@@ -1067,7 +1069,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
             :class:`~attachment.Attachment` instance.
 
         :returns: One :class:`~attachment.Attachment`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         """
         return self._get(_attachment.Attachment, attachment)
@@ -1095,7 +1097,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
             :class:`~openstack.block_storage.v3.attachment.Attachment`
             instance.
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be
+            :class:`~openstack.exceptions.NotFoundException` will be
             raised when the attachment does not exist.
             When set to ``True``, no exception will be set when
             attempting to delete a nonexistent attachment.
@@ -1197,14 +1199,14 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
 
         :param snapshot: The name or ID a backup
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the backup does not exist.
         :param bool details: When set to ``False`` no additional details will
             be returned. The default, ``True``, will cause objects with
             additional attributes to be returned.
 
         :returns: One :class:`~openstack.block_storage.v3.backup.Backup`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         :raises: :class:`~openstack.exceptions.DuplicateResource` when multiple
             resources are found.
@@ -1235,7 +1237,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
         :param backup: The value can be the ID of a backup or a
             :class:`~openstack.block_storage.v3.backup.Backup` instance
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised when
+            :class:`~openstack.exceptions.NotFoundException` will be raised when
             the zone does not exist.
             When set to ``True``, no exception will be set when attempting to
             delete a nonexistent zone.
@@ -1291,7 +1293,11 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
         params = {}
         if project:
             params['project_id'] = resource.Resource._get_id(project)
-        return self._get(_limits.Limits, requires_id=False, **params)
+
+        # we don't use Proxy._get since that doesn't allow passing arbitrary
+        # query string parameters
+        res = self._get_resource(_limits.Limits, None)
+        return res.fetch(self, requires_id=False, **params)
 
     # ====== CAPABILITIES ======
     def get_capabilities(self, host):
@@ -1301,7 +1307,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
 
         :returns: One :class:
             `~openstack.block_storage.v3.capabilites.Capabilities` instance.
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             resource can be found.
         """
         return self._get(_capabilities.Capabilities, host)
@@ -1324,14 +1330,14 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
 
         :param name_or_id: The name or ID of a group.
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the group snapshot does not exist.
         :param bool details: When set to ``False``, no additional details will
             be returned. The default, ``True``, will cause additional details
             to be returned.
 
         :returns: One :class:`~openstack.block_storage.v3.group.Group`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         :raises: :class:`~openstack.exceptions.DuplicateResource` when multiple
             resources are found.
@@ -1463,14 +1469,14 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
 
         :param name_or_id: The name or ID of a group snapshot.
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the group snapshot does not exist.
         :param bool details: When set to ``False``, no additional details will
             be returned. The default, ``True``, will cause additional details
             to be returned.
 
         :returns: One :class:`~openstack.block_storage.v3.group_snapshot`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         :raises: :class:`~openstack.exceptions.DuplicateResource` when multiple
             resources are found.
@@ -1551,7 +1557,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
 
         :returns: One :class:
             `~openstack.block_storage.v3.group_type.GroupType` instance.
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             resource can be found.
         """
         return self._get(_group_type.GroupType, group_type)
@@ -1561,12 +1567,12 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
 
         :param name_or_id: The name or ID of a group type.
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the group type does not exist.
 
         :returns: One
             :class:`~openstack.block_storage.v3.group_type.GroupType`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             resource can be found.
         :raises: :class:`~openstack.exceptions.DuplicateResource` when multiple
             resources are found.
@@ -1618,7 +1624,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
             or a :class:`~openstack.block_storage.v3.group_type.GroupType`
             instance.
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised when
+            :class:`~openstack.exceptions.NotFoundException` will be raised when
             the zone does not exist.
             When set to ``True``, no exception will be set when attempting to
             delete a nonexistent zone.
@@ -1716,7 +1722,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
 
         :returns: One
             :class:`~openstack.block_storage.v3.quota_class_set.QuotaClassSet`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         """
         return self._get(_quota_class_set.QuotaClassSet, quota_class_set)
@@ -1751,7 +1757,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
         :param dict query: Additional query parameters to use.
 
         :returns: One :class:`~openstack.block_storage.v3.quota_set.QuotaSet`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         """
         project = self._get_resource(_project.Project, project)
@@ -1768,7 +1774,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
             which the quota should be retrieved
 
         :returns: One :class:`~openstack.block_storage.v3.quota_set.QuotaSet`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         """
         project = self._get_resource(_project.Project, project)
@@ -1794,22 +1800,37 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
 
         return res.delete(self, **query)
 
-    def update_quota_set(self, quota_set, query=None, **attrs):
+    def update_quota_set(self, project, **attrs):
         """Update a QuotaSet.
 
-        :param quota_set: Either the ID of a quota_set or a
-            :class:`~openstack.block_storage.v3.quota_set.QuotaSet` instance.
-        :param dict query: Optional parameters to be used with update call.
+        :param project: ID or instance of
+            :class:`~openstack.identity.project.Project` of the project for
+            which the quota should be reset.
         :param attrs: The attributes to update on the QuotaSet represented
             by ``quota_set``.
 
         :returns: The updated QuotaSet
         :rtype: :class:`~openstack.block_storage.v3.quota_set.QuotaSet`
         """
-        res = self._get_resource(_quota_set.QuotaSet, quota_set, **attrs)
-        if not query:
-            query = {}
-        return res.commit(self, **query)
+        if 'project_id' in attrs or isinstance(project, _quota_set.QuotaSet):
+            warnings.warn(
+                "The signature of 'update_quota_set' has changed and it "
+                "now expects a Project as the first argument, in line "
+                "with the other quota set methods.",
+                os_warnings.RemovedInSDK50Warning,
+            )
+            if isinstance(project, _quota_set.QuotaSet):
+                attrs['project_id'] = project.project_id
+
+            # cinder doesn't support any query parameters so we simply pop
+            # these
+            if 'query' in attrs:
+                attrs.pop('params')
+        else:
+            project = self._get_resource(_project.Project, project)
+            attrs['project_id'] = project.id
+
+        return self._update(_quota_set.QuotaSet, None, **attrs)
 
     # ====== SERVICES ======
     @ty.overload
@@ -1848,14 +1869,14 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
 
         :param name_or_id: The name or ID of a service
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised when
+            :class:`~openstack.exceptions.NotFoundException` will be raised when
             the resource does not exist.
             When set to ``True``, None will be returned when attempting to find
             a nonexistent resource.
         :param dict query: Additional attributes like 'host'
 
         :returns: One: class:`~openstack.block_storage.v3.service.Service` or None
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         :raises: :class:`~openstack.exceptions.DuplicateResource` when multiple
             resources are found.
@@ -1892,8 +1913,8 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
         :returns: Updated service instance
         :rtype: class: `~openstack.block_storage.v3.service.Service`
         """
-        service = self._get_resource(_service.Service, service)
-        return service.enable(self)
+        service_obj = self._get_resource(_service.Service, service)
+        return service_obj.enable(self)
 
     def disable_service(
         self,
@@ -1910,8 +1931,8 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
         :returns: Updated service instance
         :rtype: class: `~openstack.block_storage.v3.service.Service`
         """
-        service = self._get_resource(_service.Service, service)
-        return service.disable(self, reason=reason)
+        service_obj = self._get_resource(_service.Service, service)
+        return service_obj.disable(self, reason=reason)
 
     def thaw_service(
         self,
@@ -1925,8 +1946,8 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
         :returns: Updated service instance
         :rtype: class: `~openstack.block_storage.v3.service.Service`
         """
-        service = self._get_resource(_service.Service, service)
-        return service.thaw(self)
+        service_obj = self._get_resource(_service.Service, service)
+        return service_obj.thaw(self)
 
     def freeze_service(
         self,
@@ -1940,8 +1961,8 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
         :returns: Updated service instance
         :rtype: class: `~openstack.block_storage.v3.service.Service`
         """
-        service = self._get_resource(_service.Service, service)
-        return service.freeze(self)
+        service_obj = self._get_resource(_service.Service, service)
+        return service_obj.freeze(self)
 
     def failover_service(
         self,
@@ -1960,8 +1981,10 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
         :returns: Updated service instance
         :rtype: class: `~openstack.block_storage.v3.service.Service`
         """
-        service = self._get_resource(_service.Service, service)
-        return service.failover(self, cluster=cluster, backend_id=backend_id)
+        service_obj = self._get_resource(_service.Service, service)
+        return service_obj.failover(
+            self, cluster=cluster, backend_id=backend_id
+        )
 
     # ====== RESOURCE FILTERS ======
     def resource_filters(self, **query):
@@ -2001,7 +2024,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
         :param transfer: The value can be either the ID of a transfer or a
             :class:`~openstack.block_storage.v3.transfer.Transfer`` instance.
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be
+            :class:`~openstack.exceptions.NotFoundException` will be
             raised when the transfer does not exist.
             When set to ``True``, no exception will be set when
             attempting to delete a nonexistent transfer.
@@ -2019,11 +2042,11 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
 
         :param name_or_id: The name or ID a transfer
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the volume transfer does not exist.
 
         :returns: One :class:`~openstack.block_storage.v3.transfer.Transfer`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         :raises: :class:`~openstack.exceptions.DuplicateResource` when multiple
             resources are found.
@@ -2042,7 +2065,7 @@ class Proxy(_base_proxy.BaseBlockStorageProxy):
             instance.
 
         :returns: One :class:`~openstack.block_storage.v3.transfer.Transfer`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no resource can be found.
         """
         return self._get(_transfer.Transfer, transfer)
