@@ -18,6 +18,8 @@ class ServiceProfile(resource.Resource):
     resources_key = 'service_profiles'
     base_path = '/service_profiles'
 
+    _allow_unknown_attrs_in_body = True
+
     # capabilities
     allow_create = True
     allow_fetch = True
@@ -26,9 +28,10 @@ class ServiceProfile(resource.Resource):
     allow_list = True
 
     _query_mapping = resource.QueryParameters(
-        'description', 'driver',
+        'description',
+        'driver',
+        'project_id',
         is_enabled='enabled',
-        project_id='tenant_id'
     )
     # Properties
     #: Description of the service flavor profile.
@@ -40,4 +43,6 @@ class ServiceProfile(resource.Resource):
     #: Metainformation of the service flavor profile
     meta_info = resource.Body('metainfo')
     #: The owner project ID
-    project_id = resource.Body('tenant_id')
+    project_id = resource.Body('project_id', alias='tenant_id')
+    #: Tenant_id (deprecated attribute).
+    tenant_id = resource.Body('tenant_id', deprecated=True)
